#############################################################################
#O  VectorSpaceToElement( <geom>, <v> ) returns the elements in <geom> determined
# by the vectorspace <v>. Several checks are built in. 
##
InstallMethod( VectorSpaceToElement, 
	"for a projective space and a Plist",
	[IsProjectiveSpace, IsPlistRep],
	function( geom, v )
		local  x, n, i;
		## when v is empty... 
        if IsEmpty(v) then
			Error("<v> does not represent any vectorspace");
		fi;
        #we replace the following two lines by the third.
		# x := MutableCopyMat(v);
		#TriangulizeMat(x); 
		x := SemiEchelonMat(v).vectors;
        ## dimension should be correct
		if Length(v[1]) <> geom!.dimension + 1 then
			Error("Dimensions are incompatible");
		fi;
        # the following lines are not necessary any more, since zero rows are removed by SemiEchelonMat.
		# but: this gives a possibility that we end up with the empty subspace.
		## Remove zero rows. It is possible the the user
		## has inputted a matrix which does not have full rank
        #n := Length(x);
		#i := 0;
		#while i < n and ForAll(x[n-i], IsZero) do
		#	i := i+1; 
		#od;
		#if i = n then
		if Length(x) = 0 then
			return EmptySubspace(geom);
		fi;
		#x := x{[1..n-i]};
		if Length(x)=ProjectiveDimension(geom)+1 then
			return geom;
		fi;
        ## It is possible that (a) the user has entered a
		## matrix with one row, or that (b) the user has
		## entered a matrix with rank 1 (thus at this stage
		## we will have a matrix with one row).
        ## We must also compress our vector/matrix.
        if Length(x) = 1 then
			x := x[1];
			ConvertToMatrixRep(x, geom!.basefield); # the extra basefield is necessary.
			return Wrap(geom, 1, x);
		else
			ConvertToMatrixRep(x, geom!.basefield); # the same here.
			return Wrap(geom, Length(x), x);
		fi;
	end );

# CHECKED 8/09/11 jdb
# CHANGED 19/9/2011 jdb + ml
#############################################################################
#O  VectorSpaceToElement( <geom>, <v> ) returns the elements in <geom> determined
# by the vectorspace <v>. Several checks are built in. 
##
InstallMethod( VectorSpaceToElement, 
	"for a compressed GF(2)-matrix",
	[IsProjectiveSpace, IsGF2MatrixRep],
	function( geom, v )
		local  x, n, i;
		## when v is empty... 
		if IsEmpty(v) then
			Error("<v> does not represent any vectorspace");
		fi;
        #we replace the following two lines by the third.
		# x := MutableCopyMat(v);
		#TriangulizeMat(x); 
		x := SemiEchelonMat(v).vectors;
		## dimension should be correct
		if Length(v[1]) <> geom!.dimension + 1 then
			Error("Dimensions are incompatible");
		fi;
		## Remove zero rows. It is possible the the user
		## has inputted a matrix which does not have full rank
		n := Length(x);
		i := 0;
		while i < n and ForAll(x[n-i], IsZero) do
			i := i+1; 
		od;
		if i = n then
			return EmptySubspace(geom);
		fi;
		x := x{[1..n-i]};
		if Length(x)=ProjectiveDimension(geom)+1 then
			return geom;
		fi;
		## It is possible that (a) the user has entered a
		## matrix with one row, or that (b) the user has
		## entered a matrix with rank 1 (thus at this stage
		## we will have a matrix with one row).
	    ## We must also compress our vector/matrix.
		if Length(x) = 1 then
			ConvertToVectorRep(x, geom!.basefield);
			return Wrap(geom, 1, x[1]);
		else
			ConvertToMatrixRep(x, geom!.basefield);
			return Wrap(geom, Length(x), x);
		fi;
	end );
  
# CHECKED 8/09/11 jdb
#############################################################################
#O  VectorSpaceToElement( <geom>, <v> ) returns the elements in <geom> determined
# by the vectorspace <v>. Several checks are built in. 
##
InstallMethod( VectorSpaceToElement, 
	"for a compressed basis of a vector subspace",
	[IsProjectiveSpace, Is8BitMatrixRep],
	function( geom, v )
		local  x, n, i;
		## when v is empty... 
		if IsEmpty(v) then
			Error("<v> does not represent any vectorspace");
		fi;
		# x := MutableCopyMat(v);
		#TriangulizeMat(x); 
		x := SemiEchelonMat(v).vectors;
		## dimension should be correct
		if Length(v[1]) <> geom!.dimension + 1 then
			Error("Dimensions are incompatible");
		fi;	
		## Remove zero rows. It is possible the the user
		## has inputted a matrix which does not have full rank
		n := Length(x);
		i := 0;
		while i < n and ForAll(x[n-i], IsZero) do
			i := i+1; 
		od;
		if i = n then
			return EmptySubspace(geom);
		fi;
		x := x{[1..n-i]};
		if Length(x)=ProjectiveDimension(geom)+1 then
			return geom;
		fi;
		## It is possible that (a) the user has entered a
		## matrix with one row, or that (b) the user has
		## entered a matrix with rank 1 (thus at this stage
		## we will have a matrix with one row).
		## We must also compress our vector/matrix.
		if Length(x) = 1 then
			ConvertToVectorRep(x, geom!.basefield);
			return Wrap(geom, 1, x[1]);
		else
			ConvertToMatrixRep(x, geom!.basefield);
			return Wrap(geom, Length(x), x);
		fi;
	end );  
  
# CHECKED 11/04/15 jdb
#############################################################################
#O  VectorSpaceToElement( <geom>, <v> ) returns the elements in <geom> determined
# by the rowvector <v>. Several checks are built in.
##
InstallMethod( VectorSpaceToElement,
	"for a row vector",
	[IsProjectiveSpace, IsRowVector],
	function( geom, v )
		local  x, n, i;
		## when v is empty...
		if IsEmpty(v) then
			Error("<v> does not represent any vectorspace");
		fi;
		x := ShallowCopy(v);
		## dimension should be correct
		if Length(v) <> geom!.dimension + 1 then
			Error("Dimensions are incompatible");
		fi;
		## We must also compress our vector.
		ConvertToVectorRep(x, geom!.basefield);
		## bad characters, such as jdb, checked this with input zero vector...
		if IsZero(x) then
			return EmptySubspace(geom);
		else
			#something new here (19/9/2011)
			MultRowVector(x,Inverse( x[PositionNonZero(x)] ));
			return Wrap(geom, 1, x);
		fi;
	end );

# CHECKED 11/04/15 jdb
#############################################################################
#O  VectorSpaceToElement( <geom>, <v> ) returns the elements in <geom> determined
# by the rowvector <v>. Several checks are built in.
##
InstallMethod( VectorSpaceToElement, 
	"for an 8-bit vector",
	[IsProjectiveSpace, Is8BitVectorRep],
	function( geom, v )
		local  x, n, i;
		## when v is empty...
		if IsEmpty(v) then
			return EmptySubspace;
		fi;
		x := ShallowCopy(v);
		## dimension should be correct
		if Length(v) <> geom!.dimension + 1 then
			Error("Dimensions are incompatible");
		fi;
		## We must also compress our vector.
		ConvertToVectorRep(x, geom!.basefield);
		## bad characters, such as jdb, checked this with input zero vector...
		if IsZero(x) then
			return EmptySubspace(geom);
		else
			return Wrap(geom, 1, x);
		fi;
	end );
